// migrate all prefs to "extensions." branch
(function migratePrefsToExtensionsBranch(fromBranch) {
	var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("");
	prefs.getChildList(fromBranch, {}).forEach(function(name) {
		if (prefs.prefHasUserValue(name)) {
			switch (prefs.getPrefType(name)) {
				case prefs.PREF_BOOL:
					prefs.setBoolPref("extensions."+name, prefs.getBoolPref(name));
					break;
				case prefs.PREF_INT:
					prefs.setIntPref("extensions."+name, prefs.getIntPref(name));
					break;
				case prefs.PREF_STRING:
					prefs.setCharPref("extensions."+name, prefs.getCharPref(name));
					break;
			}
			prefs.clearUserPref(name);
		}
	});
})("hidemenubar.");

window.addEventListener("load", function(event) {
	window.removeEventListener("load", arguments.callee, false);

	var menubox = document.getElementById("menubar-items");
	var toolbar = menubox.parentNode;
	var menubar = menubox.firstChild;
	var toolbox = toolbar.parentNode;

	// handle: "autohide" property
	toolbar.__defineGetter__("autohide",
		function() {
			return (toolbar.getAttribute("autohide") == "true");
		}
	);
	toolbar.__defineSetter__("autohide",
		function() {
			if (arguments[0]) {
				toolbar.setAttribute("autohide", "true");
			} else {
				toolbar.removeAttribute("autohide");
			}
		}
	);

	// handle: "inactive" property
	toolbar.__defineGetter__("inactive",
		function() {
			return (toolbar.getAttribute("inactive") == "true");
		}
	);
	toolbar.__defineSetter__("inactive",
		function() {
			if (arguments[0]) {
				toolbar.setAttribute("inactive", "true");
			} else {
				toolbar.removeAttribute("inactive");
			}
		}
	);

	// handle: "mousehover" property
	toolbar.__defineGetter__("hasMouseHover",
		function() {
			return (toolbar.getAttribute("mousehover") == "true");
		}
	);
	toolbar.__defineSetter__("hasMouseHover",
		function() {
			if (arguments[0]) {
				toolbar.setAttribute("mousehover", "true");
			} else {
				toolbar.removeAttribute("mousehover");
			}
		}
	);

	// handle: "popupshown" property
	toolbar.__defineGetter__("hasPopup",
		function() {
			return (toolbar.getAttribute("popupshown") == "true");
		}
	);
	toolbar.__defineSetter__("hasPopup",
		function() {
			if (arguments[0]) {
				toolbar.setAttribute("popupshown", "true");
			} else {
				toolbar.removeAttribute("popupshown");
			}
		}
	);

	// handle: "focused" property
	toolbar.__defineGetter__("hasFocus",
		function() {
			return (toolbar.getAttribute("focused") == "true");
		}
	);
	toolbar.__defineSetter__("hasFocus",
		function() {
			if (arguments[0]) {
				toolbar.setAttribute("focused", "true");
			} else {
				toolbar.removeAttribute("focused");
			}
		}
	);

	// handle: "menuactive" property
	toolbar.__defineGetter__("hasMenuActive",
		function() {
			return (toolbar.getAttribute("menuactive") == "true");
		}
	);
	toolbar.__defineSetter__("hasMenuActive",
		function() {
			if (arguments[0]) {
				toolbar.setAttribute("menuactive", "true");
			} else {
				toolbar.removeAttribute("menuactive");
			}
		}
	);

	// handle: "floating" property
	toolbar.__defineGetter__("floating",
		function() {
			return nsPreferences.getBoolPref("extensions.hidemenubar.floating", true);
		}
	);

	// implement: press [ALT] key to control the menubar display
	// XXX: pressing [AltGr], the keyCode sequence is:
	//   1.keyDown: DOM_VK_CONTROL [C]
	//   2.keyDown: DOM_VK_ALT     [C] [A]
	//   3.keyUp:   DOM_VK_CONTROL [A]
	//  ** Note: the keyUp event of DOM_VK_ALT doesn't exist.
	const TOGGLE_NOTHING = 0;
	const TOGGLE_MENUBAR = 1;
	const TOGGLE_APPMENU = 2;
	// function: getPrefToggleMode
	function getPrefToggleMode() {
		var toggle = nsPreferences.getIntPref("extensions.hidemenubar.toggle.altKey", TOGGLE_MENUBAR);
		if (toggle == TOGGLE_APPMENU) {
			var titlebar = document.getElementById("titlebar");
			if (titlebar && titlebar.hidden) toggle = TOGGLE_MENUBAR;
		}
		return toggle;
	}
	// handle keyDown event
	window.addEventListener("keydown", function(event) {
		toolbar._F10Key = (event.keyCode == KeyEvent.DOM_VK_F10);
		if (!toolbar.hidden) {
			var keyDownAlt = (event.keyCode == KeyEvent.DOM_VK_ALT) && !event.ctrlKey && !event.shiftKey && !event.metaKey;
			var keyDownAltGr = ((event.keyCode == KeyEvent.DOM_VK_ALT) && event.ctrlKey) && !event.shiftKey && !event.metaKey;
			if (keyDownAlt || keyDownAltGr) {
				if (!window.document.popupNode && !toolbar._altKey) {
					switch (getPrefToggleMode()) {
					case TOGGLE_APPMENU:
						var appmenu = document.getElementById("appmenu-popup");
						if (appmenu) {
							toolbar._altKey = (appmenu.state == "closed");
							break;
						}
					case TOGGLE_MENUBAR:
						toolbar._inactive = toolbar.inactive;
						toolbar._altKey = true;
						break;
					}
				}
				return;
			}
			delete toolbar._inactive;
			delete toolbar._altKey;
		}
	}, true);
	// handle keyUp event
	window.addEventListener("keyup", function(event) {
		if (toolbar._altKey || toolbar._F10Key) toolbar.autohide = true; // if press [ALT] or [F10], force the menubar autohide
		toolbar._F10Key = false;
		if (!toolbar.hidden && toolbar._altKey) {
			var keyUpAlt = (event.keyCode == KeyEvent.DOM_VK_ALT) && !event.ctrlKey && !event.shiftKey && !event.metaKey;
			var keyUpAltGr = ((event.keyCode == KeyEvent.DOM_VK_CONTROL) && event.altKey) && !event.shiftKey && !event.metaKey;
			if (keyUpAlt || keyUpAltGr) {
				switch (getPrefToggleMode()) {
				case TOGGLE_APPMENU:
					var appmenu = document.getElementById("appmenu-popup");
					if (appmenu && appmenu.openPopup) {
						appmenu.openPopup(appmenu.parentNode, "after_start", 0, 0, false, false);
						break;
					}
				case TOGGLE_MENUBAR:
					toolbar.inactive = !toolbar._inactive;
					if (!toolbar.inactive) {
						var evt = document.createEvent("KeyboardEvent");
						evt.initKeyEvent("keypress", true, true, null, false, false, false, false, KeyEvent.DOM_VK_F10, 0);
						document.dispatchEvent(evt);
					}
					break;
				}
			}
			delete toolbar._inactive;
			delete toolbar._altKey;
		}
	}, true);

	// implement: prevent menubar collapse when mouse hover on it
	toolbar.addEventListener("mouseover", function(event) {
		toolbar.hasMouseHover = true;
	}, true);
	toolbar.addEventListener("mouseout", function(event) {
		if (!nsPreferences.getBoolPref("extensions.hidemenubar.autoshow", true)) tryHideMenubar();
		toolbar.hasMouseHover = false;
	}, true);

	// implement: prevent menubar collapse when menu popup
	function isSubMenu(m) {
		while (m = m.parentNode) {
			if (m.localName == "menupopup" || m.localName == "popup") return true;
		}
		return false;
	}
	toolbar.addEventListener("popupshown", function(event) {
		if (isSubMenu(event.target)) return;
		if (event.target.localName != "menupopup" && event.target.localName != "popup") return;
		toolbar.hasPopup = true;
	}, true);
	toolbar.addEventListener("popuphidden", function(event) {
		if (isSubMenu(event.target)) return;
		if (event.target.localName != "menupopup" && event.target.localName != "popup") return;
		tryHideMenubar();
		toolbar.hasPopup = false;
	}, true);

	// implement: prevent menubar collapse when input box was focused
	toolbar.addEventListener("focus", function(event) {
		toolbar.scrollTop = 0; // [workaround]: toolbar.scrollTop < 0 problem (caused by pressing F6 to focus on menubar input box)
		toolbar.hasFocus = true;
		toolbar.inactive = false;
	}, true);
	toolbar.addEventListener("blur", function(event) {
		tryHideMenubar();
		toolbar.hasFocus = false;
	}, true);

	// implement: prevent menubar collapse when menuitem was actived
	menubar.addEventListener("DOMMenuItemActive", function(event) {
		if (event.target.parentNode != this) return;
		toolbar.hasMenuActive = true;
	}, true);
	menubar.addEventListener("DOMMenuItemInactive", function(event) {
		if (event.target.parentNode != this) return;
		toolbar.hasMenuActive = false;
	}, true);

	// implement: display all toolbars when customizing
	toolbox.addEventListener("DOMAttrModified", function(event) {
		if (event.target != this) return;
		if (event.attrName == "customizing") {
			for (var i=0; i<toolbox.childNodes.length; i++) {
				var bar = toolbox.childNodes[i];
				if (bar.getAttribute("type") == "menubar") continue;
				if (bar.getAttribute("customizable") != "true") continue;
				if (event.newValue == "true") {
					bar.setAttribute("customizing-collapsed", bar.collapsed);
					bar.collapsed = false;
				} else {
					bar.collapsed = (bar.getAttribute("customizing-collapsed") == "true");
					bar.removeAttribute("customizing-collapsed");
				}
			}
		}
	}, true);

	// function: tryHideMenubar
	function tryHideMenubar() {
		if (toolbar._beginShow) { clearTimeout(toolbar._beginShow); delete toolbar._beginShow; }
		if (toolbar._beginHide) return;
		if (toolbar.inactive == true) return;
		if (toolbar.getAttribute("autohide") != "true") return;
		if (toolbox.getAttribute("customizing") == "true") return;
		toolbar._beginHide = setTimeout(function() {
			if (!toolbar.hasFocus && !toolbar.hasPopup && !toolbar.hasMouseHover && !toolbar.hasMenuActive) {
				toolbar.inactive = true;
			}
			clearTimeout(toolbar._beginHide);
			delete toolbar._beginHide;
		}, nsPreferences.getIntPref("extensions.hidemenubar.autoshow.delay.onhide", 0));
	}

	// function: tryShowMenubar
	function tryShowMenubar() {
		if (toolbar._beginHide) { clearTimeout(toolbar._beginHide); delete toolbar._beginHide; }
		if (toolbar._beginShow) return;
		if (toolbar.inactive == false) return;
		if (toolbar.getAttribute("autohide") != "true") return;
		if (toolbox.getAttribute("customizing") == "true") return;
		toolbar._beginShow = setTimeout(function() {
			toolbar.inactive = false;
			clearTimeout(toolbar._beginShow);
			delete toolbar._beginShow;
		}, nsPreferences.getIntPref("extensions.hidemenubar.autoshow.delay.onshow", 0));
	}

	// function: tryAutoShowHide
	function tryAutoShowHide(event) {
		if (!nsPreferences.getBoolPref("extensions.hidemenubar.autoshow", true)) return;
		var marginTop = parseInt(toolbar.style.marginTop); if (isNaN(marginTop)) marginTop = 0;
		var minY = toolbar.boxObject.screenY - marginTop - nsPreferences.getIntPref("extensions.hidemenubar.autoshow.above.pixels", 0);
		var maxY = toolbar.boxObject.screenY - marginTop + nsPreferences.getIntPref("extensions.hidemenubar.autoshow.below.pixels", 0);
		if (event.screenY >= minY && event.screenY < maxY) {
			tryShowMenubar();
		} else {
			tryHideMenubar();
		}
	}

	// implement: try to show the menubar when mouse hover on the autoshow area
	window.addEventListener("mouseout", tryAutoShowHide, true);
	window.addEventListener("mousemove", tryAutoShowHide, true);

	// implement: prevent altKey to trigger DOMMenuBarActive/DOMMenuBarInactive event
	toolbar.addEventListener("DOMMenuBarActive", function(event) {
		if (!toolbar._F10Key) event.stopPropagation();
	}, true);
	toolbar.addEventListener("DOMMenuBarInactive", function(event) {
		if (!toolbar._F10Key) event.stopPropagation();
		tryHideMenubar();
	}, true);

	// define: id for menubar-wrapper, menubar-container
	const MENUBAR_WRAPPER_ID = "menubar-wrapper";
	const MENUBAR_CONTAINER_ID = "menubar-container";

	// function: add an observer for the preference modified event
	function addPrefModifiedObserver(prefName, callback) {
		const Cc = Components.classes, Ci = Components.interfaces;
		const prefService = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefBranch2).QueryInterface(Ci.nsIPrefService);
		prefService.addObserver(prefName, { observe: callback }, false);
	}

	// function: add an observer for the attribute modified event
	function addAttrModifiedObserver(elmt, attrName, callback) {
		if (typeof(elmt)=="string") elmt = document.getElementById(elmt);
		if (elmt) {
			elmt.addEventListener("DOMAttrModified", function(event) {
				if (event.target != elmt) return;
				if (event.attrName != attrName) return;
				callback(event);
			}, true);
		}
	}

	// function: update menubar-wrapper style
	function updateMenubarWrapperStyle() {
		setTimeout(function() {
			var wrapper = document.getElementById(MENUBAR_WRAPPER_ID);
			if (wrapper) {
				var marginTop = nsPreferences.getIntPref("extensions.hidemenubar.floating.marginTop", 0);
				toolbar.style.setProperty("margin-top", marginTop+"px", "important");
				wrapper.removeAttribute("style");
				var target = toolbar;
				while (target && (target instanceof XULElement)) {
					var cs = window.getComputedStyle(target, null);
					if ((cs.MozAppearance == "menubar") || (cs.backgroundColor != "transparent") || (cs.backgroundImage != "none")) {
						wrapper.style.MozAppearance = "none";
						wrapper.style.color = cs.color;
						wrapper.style.backgroundColor = cs.backgroundColor;
						wrapper.style.backgroundImage = cs.backgroundImage;
						wrapper.style.backgroundRepeat = cs.backgroundRepeat;
						wrapper.style.backgroundPosition = "100% -"+wrapper.boxObject.y+"px";
						break;
					}
					target = target.parentNode;
				}
			} else {
				toolbar.style.removeProperty("margin-top");
			}
		}, 0);
	}

	// function: dispatch "aftercustomization" event
	function dispatchAfterCustomizationEvent() {
		var evt = document.createEvent("Events");
		evt.initEvent("aftercustomization", true, true);
		toolbox.dispatchEvent(evt);
	}

	// function: wraping or unwraping the menubar items
	function wrappingMenubarItems(wrapping) {
		var wrapper = document.getElementById(MENUBAR_WRAPPER_ID);
		if (wrapping) {
			if (!wrapper) {
				wrapper = document.createElement("hbox");
				wrapper.id = MENUBAR_WRAPPER_ID;
				var container = document.createElement("hbox");
				container.id = MENUBAR_CONTAINER_ID;
				while (toolbar && toolbar.firstChild) {
					container.appendChild(toolbar.firstChild);
				}
				toolbar.appendChild(wrapper).appendChild(container);
				toolbar.setAttribute("floating", "true");
				dispatchAfterCustomizationEvent();
			}
		} else {
			if (wrapper) {
				var container = wrapper.firstChild;
				while (container && container.firstChild) {
					toolbar.appendChild(container.firstChild);
				}
				wrapper.parentNode.removeChild(wrapper);
				toolbar.removeAttribute("floating");
				dispatchAfterCustomizationEvent();
			}
		}
		updateMenubarWrapperStyle();
	}

	// observe document.documentElement's "style" attribute for update wrapper's style
	addAttrModifiedObserver(document.documentElement, "style", updateMenubarWrapperStyle);

	// observe the preference "hidemenubar.floating" for wrapping the menubar items
	addPrefModifiedObserver("extensions.hidemenubar.floating", function() {
		wrappingMenubarItems(toolbar.floating && toolbar.autohide);
	});

	// observe menubar's "autohide" attribute for wrapping the menubar items
	addAttrModifiedObserver(toolbar, "autohide", function() {
		wrappingMenubarItems(toolbar.floating && toolbar.autohide);
	});

	// observe cmd_CustomizeToolbars's "disabled" attribute for wrapping the menubar items
	addAttrModifiedObserver(toolbar.getAttribute("cmdcustomize"), "disabled", function(event) {
		wrappingMenubarItems(toolbar.floating && toolbar.autohide && (event.newValue != "true"));
	});

	// on startup, wrapping the menubar-items
	wrappingMenubarItems(toolbar.floating);

	// on startup, force the menubar autohide
	toolbar.autohide = true;

}, false);
